﻿using CPF;
using CPF.Animation;
using CPF.Charts;
using CPF.Controls;
using CPF.Drawing;
using CPF.Shapes;
using CPF.Styling;
using CPF.Svg;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CPF.Skia;
using Core.Helpers;
using Silk.NET.Maths;
using Silk.NET.OpenGL;
using Core.Models;
using System.IO;
using Shader = Core.Helpers.Shader;
using CPF.Input;
using CPF.Windows.OpenGL;
using System.Diagnostics;

namespace CPF_OpenGL
{
    public class Window2_1_Colors : Window
    {
        private static GL gl = null!;
        private static Camera camera = null!;

        #region Input
        private static bool firstMove = true;
        private static Vector2D<float> lastPos;
        #endregion

        #region Models
        private static Cube lightCube = null!;
        private static Cube lightingCube = null!;
        #endregion

        #region Positions
        private static Vector3D<float> lightPos = new(1.2f, -1.2f, -2.0f);
        private static Vector3D<float> lightingPos = new(0.0f, 0.0f, -1.0f);
        #endregion

        #region Programs
        // 光源
        private static ShaderProgram lightProgram = null!;

        // 进行光照
        private static ShaderProgram lightingProgram = null!;
        #endregion
        protected override void InitializeComponent()
        {
            LoadStyleFile("res://CPF_OpenGL/Stylesheet1.css");
            //加载样式文件，文件需要设置为内嵌资源

            Title = "OpenGL测试";
            Width = 500;
            Height = 400;
            Background = null;
            CanResize = true;
            Children.Add(new WindowFrame(this, new GLView
            {
                Width = "100%",
                Height = "100%",
                [nameof(GLView.GLLoaded)] = new CommandDescribe((s, e) =>
                {
                    gl = GL.GetApi(((GLEventArgs)e).Context.GetProcAddress);
                    camera = new Camera();
                    lightCube = new Cube(gl);
                    lightingCube = new Cube(gl);
                    using Shader mvp = new(gl, GLEnum.VertexShader, ResourceManager.GetText("res://CPF_OpenGL/S2_1_Colors/Shaders/mvp.vert").Result);
                    using Shader light = new(gl, GLEnum.FragmentShader, ResourceManager.GetText("res://CPF_OpenGL/S2_1_Colors/Shaders/light.frag").Result);
                    using Shader lighting = new(gl, GLEnum.FragmentShader, ResourceManager.GetText("res://CPF_OpenGL/S2_1_Colors/Shaders/lighting.frag").Result);
                    lightProgram = new ShaderProgram(gl);
                    lightProgram.Attach(mvp, light);
                    lightingProgram = new ShaderProgram(gl);
                    lightingProgram.Attach(mvp, lighting);
                    gl.Enable(GLEnum.DepthTest);
                }),
                [nameof(GLView.GLRender)] = new CommandDescribe((s, e) => GLRender((GLView)s, (GLEventArgs)e)),
                [nameof(GLView.MouseDown)] = new CommandDescribe((s, e) => MouseEvent((GLView)s, (MouseEventArgs)e)),
                [nameof(GLView.MouseUp)] = new CommandDescribe((s, e) => MouseEvent((GLView)s, (MouseEventArgs)e)),
                [nameof(GLView.MouseMove)] = new CommandDescribe((s, e) => MouseEvent((GLView)s, (MouseEventArgs)e)),
            })
            {
                MaximizeBox = true
            });
            if (!DesignMode)//设计模式下不执行，也可以用#if !DesignMode
            {
                
            }
            //Children.Add(new Button
            //{
            //    MarginLeft = 199,
            //    MarginTop = 207,
            //    Height = 32,
            //    Width = 102,
            //    Content = "OpenGL版本：" + CPF.Windows.OpenGL.Wgl.VersionString,
            //});
        }

#if !DesignMode //用户代码写到这里，设计器下不执行，防止设计器出错
        protected override void OnInitialized()
        {
            base.OnInitialized();
            timer.Start();
            timer.Tick += Timer_Tick;
        }

        private void Timer_Tick(object sender, EventArgs e)
        {//实时刷新
            Invalidate();
        }

        //可以用定时器来刷新，或者用操作事件刷新
        CPF.Threading.DispatcherTimer timer = new CPF.Threading.DispatcherTimer { Interval = TimeSpan.FromMilliseconds(20) };

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);
            timer.Dispose();
        }
#endif

        void MouseEvent(GLView view, MouseEventArgs args)
        {
            if (args.LeftButton == MouseButtonState.Pressed)
            {
                Vector2D<float> vector = new(args.Location.X, args.Location.Y);

                if (firstMove)
                {
                    lastPos = vector;

                    firstMove = false;
                }
                else
                {
                    float deltaX = vector.X - lastPos.X;
                    float deltaY = vector.Y - lastPos.Y;

                    camera.Yaw += deltaX * 0.2f;
                    camera.Pitch += -deltaY * 0.2f;

                    lastPos = vector;
                }
            }
            else
            {
                firstMove = true;
            }
        }

        void GLRender(GLView view, GLEventArgs args)
        {
            if (lightProgram == null)
            {
                return;
            }
            gl.Enable(GLEnum.DepthTest);
            gl.DepthMask(true);

            gl.ClearColor(0, 0, 0, .3f);
            camera.Width = (int)(view.ActualSize.Width * RenderScaling);
            camera.Height = (int)(view.ActualSize.Height * RenderScaling);

            lightCube.Transform = Matrix4X4.CreateScale(new Vector3D<float>(0.5f, 0.5f, 0.5f)) * Matrix4X4.CreateTranslation(lightPos);
            lightingCube.Transform = Matrix4X4.CreateTranslation(lightingPos);

            gl.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit | ClearBufferMask.StencilBufferBit);

            // 光源
            {
                uint positionAttrib = (uint)lightProgram.GetAttrib("position");

                gl.EnableVertexAttribArray(positionAttrib);

                lightProgram.Enable();

                lightProgram.SetUniform("model", lightCube.Transform);
                lightProgram.SetUniform("view", camera.View);
                lightProgram.SetUniform("projection", camera.Projection);

                lightCube.Draw(positionAttrib);

                lightProgram.Disable();

                gl.DisableVertexAttribArray(positionAttrib);
            }
            // 进行光照
            {
                uint positionAttrib = (uint)lightingProgram.GetAttrib("position");

                gl.EnableVertexAttribArray(positionAttrib);

                lightingProgram.Enable();

                lightingProgram.SetUniform("model", lightingCube.Transform);
                lightingProgram.SetUniform("view", camera.View);
                lightingProgram.SetUniform("projection", camera.Projection);

                lightingProgram.SetUniform("objectColor", new Vector3D<float>(1.0f, 0.5f, 0.31f));
                lightingProgram.SetUniform("lightColor", new Vector3D<float>(1.0f, 1.0f, 1.0f));

                lightingCube.Draw(positionAttrib);

                lightingProgram.Disable();

                gl.DisableVertexAttribArray(positionAttrib);
            }


            gl.Disable(GLEnum.DepthTest);

        }
    }
}
